local mod = ...


-- *************************************************************************************************************
-- Ruin Manager component -------------------------------------------------------------------------------------
-- Creates the ruins on the map the first time it is loaded
-- *************************************************************************************************************

local COMP_RUIN_MANAGER = {
	TypeName = "COMP_RUIN_MANAGER",
	ParentType = "COMPONENT",
	Properties = {
		{ Name = "InitFlag", Type = "boolean", Default = false, Flags = { "SAVE_GAME" } },
    },
}


-- onEnabled
function COMP_RUIN_MANAGER:onEnabled()

    if not self.InitFlag then

        -- randomise random number seed
        -- local nothing = math.randomseed() -- this function does not seem to be available in Foundation
        local nothing = nil
        local t = self:getLevel():getDeltaTime()
        mod:log("Randomising based on deltaTime of " .. tostring(t))
        for randA = 1, t * 100000 do
            nothing = math.random()
        end

        -- -- Create a test ruin
        -- local posRuin = {480, 0, 470}
        -- local objRuin = self:getLevel():createObject("PREFAB_RUIN_RUIN2", posRuin)

        -- -- Spawn random ruins
        self:spawnRuin(nil, nil, "PREFAB_RUIN_RUIN1", 1, 30, 0.4, 6, true, 1, nil)
        self:spawnRuin(nil, nil, "PREFAB_RUIN_RUIN2", 1, 20, 0.3, 6, true, 1, nil)
        self:spawnRuin(nil, nil, "PREFAB_RUIN_RUIN8", 1, 20, 0.4, 6, true, 1, nil)
        -- self:spawnRuin(nil, nil, "PREFAB_RUIN_RUIN7", 1, 50, 0.2, 25, true, 0, nil)

        self:spawnCollection("PREFAB_RUIN_COLLECTION6", 1, 100, 0.3, 60)
        self:spawnCollection("PREFAB_RUIN_COLLECTION3", 1, 100, 0.5, 60)
        self:spawnCollection("PREFAB_RUIN_COLLECTION5", 1, 100, 0.5, 60)

        self:spawnCollection("PREFAB_RUIN_COLLECTION1", 1, 60, 0.5, 20)
        self:spawnCollection("PREFAB_RUIN_COLLECTION2", 1, 60, 0.5, 20)
        self:spawnCollection("PREFAB_RUIN_COLLECTION4", 1, 60, 0.5, 30)


        -- Do not repeat this code
        self.InitFlag = true

    end

end


-- Spawn ruin
function COMP_RUIN_MANAGER:spawnRuin(givenX, givenY, prefab, spawnCount, minDist, maxSlope, radius, enforceProximity, rotationMode, rotationGiven)

    for i = 1, spawnCount do

        local pos = self:getSpawnPosition(givenX, givenY, minDist, maxSlope, radius, enforceProximity)
        if pos then

            self:getLevel():createObject(prefab, pos, {0, 0, 0, 1}, function(objRuin)

                -- Generate and store a random rotation angle
                -- Mode zero is no rotation
                -- Mode 1 is random rotation
                -- Mode 2 is only 180 degree rotation
                local compRuin = objRuin:getOrCreateComponent("COMP_RUIN")

                local rotation = 0
                if rotationGiven then
                    rotation = rotationGiven
                elseif rotationMode == 1 then
                    rotation = (2 * math.pi * math.random())
                elseif compRuin.RotationMode == 2 then
                    rotation = (math.random(1, 2) * math.pi)
                end

                -- mod:log(rotation)
                compRuin.Rotation = rotation

                -- Rotate the ruin (gather points and stones)
                objRuin:setRotationY(rotation)

                -- Place the visual prefab
                compRuin.StepCount = math.random(math.floor((compRuin.StepCount * 0.5) + 0.5), compRuin.StepCount)
                self:getLevel():createObject(compRuin.PrefabName .. "_PIECE_" .. string.sub("000" .. compRuin.StepCount, -3), objRuin:getGlobalPosition(), {0, 0, 0, 1}, function(objSpawned)

                    objSpawned:setRotationY(rotation)

                    -- Add grounding if it doesn't exist yet
                    local compGround = objSpawned:getOrCreateComponent("COMP_GROUNDED")
                    compGround.AutoDisable = true

                    -- Record the spawned prefab
                    compRuin.SpawnedObject = objSpawned:getId()

                end)

            end)

        end

    end

end


-- Spawn collection
function COMP_RUIN_MANAGER:spawnCollection(prefabCollection, spawnCount, minDist, maxSlope, radius)

    for i = 1, spawnCount do

        local posCollection = self:getSpawnPosition(nil, nil, minDist, maxSlope, radius, true)

        if posCollection then

            -- Spawn the collection object
            self:getLevel():createObject(prefabCollection, posCollection, {0, 0, 0, 1}, function(objCollection)

                -- Rotate the entire collection
                local rotationCollection = (2 * math.pi * math.random())
                objCollection:setRotationY(rotationCollection)

                -- For each child item within the collection
                objCollection:forEachChild(function(objChild)

                    local posChildUntested = objChild:getGlobalPosition()
                    local posChild = self:getSpawnPosition(posChildUntested[1], posChildUntested[3], 0, maxSlope, 5, false)
                    local orientChild = quaternion.getEulerAngles(objChild:getGlobalOrientation())
                    local prefabChild = string.sub(objChild.Name, 6, string.len(objChild.Name) - 4)

                    if posChild then
                        self:spawnRuin(posChild[1], posChild[3], prefabChild, 1, 0, maxSlope, 5, false, 2, orientChild[2])
                    end

                end)

            end)

        end

    end

end



-- Get Random Position -------------------------------------------------------------------------------------
function COMP_RUIN_MANAGER:getSpawnPosition(givenX, givenY, minDist, maxSlope, radius, enforceProximity)

    -- Try 50 times; it may be impossible to find a good position in some cases (e.g. player has bought and built on all hexes)
    local tryCount = 50
    if givenX and givenY then tryCount = 1 end

    for i = 1, tryCount do

        -- Pick a random position, not too close to the map edges
        local pos = {givenX, 0, givenY}
        if not givenX and not givenY then
            pos[1] = math.random(20, 980)
            pos[3] = math.random(20, 980)
        end

        -- Test for water and elevation
        local foundWater = false
        local step = 10
        if radius < 5 then step = radius * 2 end

        for x = pos[1] - radius, pos[1] + radius, step do

            for y = pos[3] - radius, pos[3] + radius, step do

                local elevationWater = self:getWaterElevation({x, pos[2], y})
                local elevationGround = self:getGroundElevation({x, pos[2], y})

                if (elevationWater > (elevationGround - 2.5)) or (elevationGround == -9999) then
                    foundWater = true
                end

            end
        end

        if foundWater then

        else

            -- Test for steepness
            local maxSlopeFound = 0
            local testSlope = 0
            local testPointA = nil
            local testPointB = nil

            testPointA = self:getGroundElevation({pos[1], pos[2], pos[3] + radius})
            testPointB = self:getGroundElevation({pos[1], pos[2], pos[3] - radius})
            testSlope = math.abs(testPointA - testPointB) / (radius * 2) -- rise over run
            if testSlope > maxSlopeFound then
                maxSlopeFound = testSlope
            end

            testPointA = self:getGroundElevation({pos[1] + radius, pos[2], pos[3]})
            testPointB = self:getGroundElevation({pos[1] - radius, pos[2], pos[3]})
            testSlope = math.abs(testPointA - testPointB) / (radius * 2) -- rise over run
            if testSlope > maxSlopeFound then
                maxSlopeFound = testSlope
            end

            if maxSlopeFound > maxSlope then

            else

                -- test that the location is at least `minDist` metres from an existing building part
                local foundObject = false
                if enforceProximity then

                    self:getLevel():getComponentManager("COMP_WORKPLACE"):getAllEnabledComponent():forEach(function(compWork)
                        local posPart = compWork:getOwner():getGlobalPosition()
                        local dist = math.sqrt( (pos[1] - posPart.x)^2 + (pos[3] - posPart.z)^2 )
                        if (dist < minDist) then
                            foundObject = true
                        end
                    end)

                    -- test that the location is at least 'mindist' metres from another ruin
                    self:getLevel():getComponentManager("COMP_RUIN"):getAllEnabledComponent():forEach(function(compRuin)
                        local posRuin = compRuin:getOwner():getGlobalPosition()
                        local dist = math.sqrt( (pos[1] - posRuin.x)^2 + (pos[3] - posRuin.z)^2 )
                        if (dist < minDist) then
                            foundObject = true
                        end
                    end)

                end

                if not foundObject then
                    return pos
                end

            end

        end

    end

    return nil

end



-- Get Water Elevation -------------------------------------------------------------------------------------
function COMP_RUIN_MANAGER:getWaterElevation(pos)

    local raycastResult = {}
    local fromPosition = { pos[1], pos[2]+1000, pos[3] }
    local toPosition = { pos[1], pos[2]-1000, pos[3] }
    if not self:getLevel():rayCast(fromPosition, toPosition, raycastResult, 2 ^ OBJECT_FLAG.WATER:toNumber()) then
        return -9999
    else
        return raycastResult["Position"][2]
    end

end



-- Get Ground Elevation -------------------------------------------------------------------------------------
function COMP_RUIN_MANAGER:getGroundElevation(pos)

    local raycastResult = {}
    local fromPosition = { pos[1], pos[2]+1000, pos[3] }
    local toPosition = { pos[1], pos[2]-1000, pos[3] }
    if not self:getLevel():rayCast(fromPosition, toPosition, raycastResult, 2 ^ OBJECT_FLAG.TERRAIN:toNumber()) then
        return -9999
    else
        return raycastResult["Position"][2]
    end

end



-- Register custom Class
mod:registerClass(COMP_RUIN_MANAGER)


-- Attach class to prefab
mod:registerPrefabComponent("PREFAB_MANAGER", { DataType = "COMP_RUIN_MANAGER" } )



